/*****************************************************************************
 * ---------------------------------------------------------------------------
 *	S3 SDAC/Trio32/Trio64 clock chip driver
 * ---------------------------------------------------------------------------
 *
 *	Copyright (C) 1995 by Steffen Seeger [seeger@physik.tu-chemnitz.de].
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; see the file COPYING.  If not, write to
 *	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * ---------------------------------------------------------------------------
 *	This code was derived from the following source of information:
 *	[1]	80C716 SDAC Data Book (c) 1994 by S3 Incorporated
 *	[2]	Trio32/Trio64 Data Book (c) 1994 by S3 Incorporated
 * 
 ****************************************************************************/

#define	MAINTAINER	"Jon Taylor <taylorj@gaia.ecs.csus.edu>"

#include <linux/kernel.h>

#include "graphdev.h"
#include "kgi.h"

#include "S3/trio64.h"

static struct kgi_clock clock = {
  "S3 Incorporated",	/* manufacturer		*/
  "Trio32/Trio64",	/* model		*/
  GM_ALL,			/* modes		*/ 
  0,			/* flags		*/
  { 0x7FFF, 0x7FFF },	/* maximum resolution	*/
  CLK_TYPE_PROG,		/* clock type		*/
  {
      { 
	20000000, 135000000,	/* clock ranges		*/
	  0,         0,
	  0,         0,
	  0,         0,
	  0,         0,
	  0,         0,
	  0,         0,
	  0,         0
      }
  }
};

#define	MIN_FREQ clock.clock.range[0].min
#define	MAX_FREQ clock.clock.range[0].max

/* Clock driver dependend #defines */

#define CLK_VERSION "0.01 ALPHA"
#define	fref	14318180   /* Hz, must be within 14..15 MHz */ 
#define F(M,N,FVCO)  (((M)+2)*(FVCO)/((N)+2))
#define	INVALIDCLK 0xFFFF

static unsigned short clk_last_cmd = 0xFFFF;
static int 	      clk_last_freq = -1;

/* void setPLL(unsigned short cmd, int clk) {
  outbSR (0x12, clk >> 8);
  outbSR (0x13, clk & 0xff);
} */

unsigned short calcfreq(int *freq, enum kgi_tm_cmd cmd) {
  int fvco = *freq;
  int hf   = *freq;
  int div  = 0;
  int n;
  struct { unsigned char m; unsigned char n; int delta; }
         best = { 0xFF, 0xFF, 0x7FFFFFFF };    
 
  if(*freq == clk_last_freq)	/* we never get called with freq == -1 */
    return clk_last_cmd;
  
  while((div < 4) && (hf < 135000000))
    { div += 1; hf += hf; }
 
  if((hf < 135000000) || (270000000 < hf) || (div > 3))
    {
      TRACE(printk("%s:%i: invalid clock request (%i Hz)\n",__FILE__,__LINE__,\
                   *freq));
      return INVALIDCLK;
    }

  fvco  = fref >> div;

  for(n = 1; n < 32; n++)
    { 
      register int m = hf*(n+2)/fref - 2;
      register int delta = *freq - F(m,n,fvco);

      delta = (cmd == CMD_LOWER) ? delta : -delta;

      if(m > 127) break;
      if((delta < 0) || (delta > best.delta)) continue;

      best.n = n; best.m = m; best.delta = delta;
    }

  if((best.n > 31) || (best.m > 127) || (div > 3))
    {
      TRACE(printk("%s:%i: invalid clock request (%i Hz), n=%i, m=%i, d=%i\n",
                   __FILE__,__LINE__, *freq, best.n, best.m, div)); 
      return INVALIDCLK; 
    }
   else
    {
      TRACE(printk("%s: request for DCLK = %i Hz,", __FILE__,*freq ));
      *freq = clk_last_freq = F(best.m,best.n,fref >> div);
      TRACE(printk(" did %i Hz.\n", *freq));
      return clk_last_cmd = best.m | (best.n << 8) | (div << 13);
    }
}

/* GGI clock driver interface */

static unsigned short PLL_save[4];
static unsigned char clk, PLL_ctrl;

int kgi_clock_init(void) {
  /* If a Trio64 chipset was detected, this *must* be present */
  return EOK;
}

void kgi_clock_done(void) {
  TRACE(printk("%s %s clock driver removed.\n", clock.manufact, clock.model));
}

struct kgi_clock *kgi_clock_get_info (void){ 
  return &clock;
}

int kgi_clock_check_mode(struct kgi_mode *mode, enum kgi_tm_cmd cmd) {
  int newfreq = (cmd == CMD_PROPOSE) 
                    ? 0 : mode->dclk * mode->clk.mul / mode->clk.div;

  switch (cmd) {
   case CMD_PROPOSE:
    mode->clock = INVALIDCLK;
    return EOK;

   case CMD_LOWER:
    if (newfreq < MIN_FREQ) {
      TRACE(printk("%s: requested DCLK too low (%i Hz).\n", __FILE__, mode->dclk));
      return -ECLOCK;
    }

    if (newfreq > MAX_FREQ) newfreq = MAX_FREQ;	/* meet limits */

    mode->clock = calcfreq(&newfreq, cmd);
    mode->dclk = newfreq * mode->clk.div / mode->clk.mul;

    return (mode->clock == INVALIDCLK) ? -ECLOCK : EOK;
			
   case CMD_RAISE:
    if (MAX_FREQ < newfreq) {
      TRACE(printk("%s: requested DCLK too high (%i Hz).\n", __FILE__, mode->dclk));
      return -ECLOCK;
    }

    if (newfreq < MIN_FREQ) newfreq = MIN_FREQ;

    mode->clock = calcfreq(&newfreq, cmd);
    mode->dclk = newfreq * mode->clk.div / mode->clk.mul;

    return (mode->clock == INVALIDCLK) ? -ECLOCK : EOK;

   case CMD_CHECK:
    if ((newfreq < MIN_FREQ) || (newfreq > MAX_FREQ)) return -ECLOCK;

    mode->clock = calcfreq(&newfreq, CMD_LOWER);
    
    if (   (mode->clock != INVALIDCLK)
	&& (   ((mode->dclk*mode->clk.mul/mode->clk.div) - newfreq) 
	    < ((mode->dclk*mode->clk.mul/mode->clk.div)*1/100)))
      return EOK;

    return -ECLOCK;

   default:
    INTERNALERROR;
    return -ECLOCK;
  }
}

void kgi_clock_set_mode(struct kgi_mode *mode) {
/*  if(mode->clock != INVALIDCLK) {
    outbSEQ (0x12, clk >> 8);
    outbSEQ (0x13, clk & 0xff);
  } else INTERNALERROR;  */
}

int kgi_clock_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	       unsigned long arg) {
  return -ECLOCK;
}
